/**@name LIBRARY: unSystemIntegrity_CMW.ctl

@author: Herve Milcent (AB-CO)

Creation Date: 20/11/2006

Modification History: 
  03/10/2011: Herve
  - IS-600  unLHCServices  system name and component version added in the MessageText log and in the diagnostic 
 
  16/06/2011: Herve
  - IS-547: get the list of systemAlarm pattern

  23/08/2010: Herve
    - IS-391: extended diagnostic and periodic GQ (General query: periodic readout of all the subscription): periodicity configurable 
    
	19/03/2007: Herve
		- GQ Allowed
		
	06/03/2007: Herve
		- GQ: support for distrib.
		
	02/03/2007: Herve
		- add GQ
		
version 1.0

External Functions: 
	
Internal Functions: 
	
Purpose: 
	Library of the CMW component used by the systemIntegrity.
	 	
Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. global variable: the following variables are global to the script
	. data point type needed: _UnSystemIntegrity
	. data point: CMW_systemIntegrity
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/

//--------------------------------------------------------------------------------------------------------------------------------
// constant declaration
//------------------
const string c_CMW_FRONTEND_UNSYSTEMALARM = "_unSystemAlarm";
const string c_unSystemIntegrity_CMW = "CMW_";
const string UN_SYSTEM_INTEGRITY_CMW = "CMW";
const string UN_SYSTEM_INTEGRITY_CMW_check = "unSystemIntegrity_CMWCheck";
const string c_unSystemIntegrity_CMW_Time_callback = "unSystemIntegrity_CMWTimeCallback";
const int c_unSystemIntegrity_defaultCMWCheckingDelay = 10; // default value for the checking in sec.
const int C_CMW_WAIT_BEFORE_GQ = 60;
const int C_CMW_NO_THREAD = -1;
const int C_CMW_NB_ITERATION_FOR_GQ = 60;

// 
//------------------

// global declaration
global dyn_string g_unSystemIntegrity_CMW_DSList; // list of the CMW-DS dp checked
global dyn_int g_unSystemIntegrity_CMW_DSThreadId; // list of the CMW-DS thread Id checking the counter
global int g_unSystemIntegrity_CMWCheckingDelay;
global bool g_bCMWNotStarted=true;
mapping g_munSystemIntegrity_TimeAlarm; // variable used to keep the latest alarm.
global mapping g_m_iCMWDriver_GQThreadId;
global int g_unSystemIntegrity_CMWWaitBeforeGQ;
global mapping g_bunSystemIntegrity_driverStarted;
global mapping g_m_iGQAllowed;
global dyn_int g_CMW_diDriver;
global int g_CMWSystemIntegrity_GeneralQueryThreadId;
global int g_iCMWNB_ITERATION_FOR_GQ;

//@{

//------------------------------------------------------------------------------------------------------------------------
// CMW_systemIntegrityInfo
/** Return the list of systemAlarm pattern. 
  
@par Constraints
  None

@par Usage
  Public

@par PVSS managers
  CTRL

@return list of systemAlarm pattern
*/
dyn_string CMW_systemIntegrityInfo()
{
  return makeDynString(DS_pattern,DS_Time_pattern);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_Initialize
/**
Purpose:
Get the list of CMW check by the systemIntegrity. The ones that are enabled.

@param dsResult: dyn_string, output, the list of enabled CMW that are checked

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_Initialize(dyn_string &dsResult)
{
	dyn_string dsList;
	int len, i;
	string dpToCheck;
	bool enabled;

        g_iCMWNB_ITERATION_FOR_GQ = C_CMW_NB_ITERATION_FOR_GQ;
	dpGet(UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".config.data", dsList);
	unSystemIntegrity_CMW_DataCallback("", dsList);

	dsList = dpNames(c_unSystemAlarm_dpPattern+DS_pattern+"*", c_unSystemAlarm_dpType);
	len = dynlen(dsList);
	for(i = 1; i<=len; i++) {
		dpToCheck = dpSubStr(dsList[i], DPSUB_DP);
		dpToCheck = substr(dpToCheck, strlen(c_unSystemAlarm_dpPattern+DS_pattern),strlen(dpToCheck));
		if(dpExists(dpToCheck)) {
			if(dpTypeName(dpToCheck) == UNCMW_DPTYPE) {
				dpGet(dsList[i]+".enabled", enabled);
				if(enabled)
					dynAppend(dsResult, dpToCheck);
				unSystemIntegrity_CMW_InitializeGQThread(dpToCheck);
			}
		}
	}
//DebugN("CMW_DS:", dsList, dsResult);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_InitializeGQThread
/**
Purpose:
Set the GQ thread Id for of the corresponding driver to NONE.

@param dsResult: dyn_string, output, the list of enabled CMW that are checked

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_InitializeGQThread(string sDp)
{
	int iDrv;
	
	dpGet(sDp+".communication.driver_num", iDrv);
	g_m_iCMWDriver_GQThreadId[iDrv] = C_CMW_NO_THREAD;
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_InitializeGQThread", sDp);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_GQ
/**
Purpose:
Set the GQ thread Id for of the corresponding driver to NONE.

@param dsResult: dyn_string, output, the list of enabled CMW that are checked

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_GQ(int iDriverNumber, int iDelay)
{
	bool bGQ;
//DebugTN("unSystemIntegrity_CMW_GQ", iDriverNumber, iDelay);
	if(mappingHasKey(g_m_iGQAllowed,iDriverNumber)) 
		bGQ = g_m_iGQAllowed[iDriverNumber];
	else
		bGQ = true;
	if(bGQ) {
		if(g_m_iCMWDriver_GQThreadId[iDriverNumber] == C_CMW_NO_THREAD)
			g_m_iCMWDriver_GQThreadId[iDriverNumber] = startThread("unSystemIntegrity_CMW_StartGQ", iDriverNumber, iDelay);
	}
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_GQ", iDriverNumber, iDelay, bGQ, g_m_iCMWDriver_GQThreadId);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_StartGQ
/**
Purpose:
Start the GQ thread Id for of the corresponding driver to NONE.

@param dsResult: dyn_string, output, the list of enabled CMW that are checked

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_StartGQ(int iDriverNumber, int iDelay)
{
//DebugTN("unSystemIntegrity_CMW_StartGQ", iDriverNumber, iDelay);
	delay(iDelay);
	unSystemIntegrity_CMW_GeneralQuery(getSystemName(), iDriverNumber);
	g_m_iCMWDriver_GQThreadId[iDriverNumber] = C_CMW_NO_THREAD;
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_StartGQ", iDriverNumber, iDelay, g_m_iCMWDriver_GQThreadId);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_DataCallback
/**
Purpose:
handle any systemIntegrity command.

@param sDpe1: string, input, dpe
@param sConfigData: string, input, the config data

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_DataCallback(string sDpe1, dyn_string dsConfigData)
{
	dyn_string dsTemp=dsConfigData;

	while(dynlen(dsTemp) <2)
		dynAppend(dsTemp, "0");
        if(dynlen(dsTemp)<3)
          dynAppend(dsTemp, C_CMW_NB_ITERATION_FOR_GQ);

	g_unSystemIntegrity_CMWCheckingDelay = dsTemp[1];
	if(g_unSystemIntegrity_CMWCheckingDelay <= 0)
		g_unSystemIntegrity_CMWCheckingDelay = c_unSystemIntegrity_defaultCMWCheckingDelay;

	g_unSystemIntegrity_CMWWaitBeforeGQ = dsTemp[2];
	if(g_unSystemIntegrity_CMWWaitBeforeGQ <= C_CMW_WAIT_BEFORE_GQ)
		g_unSystemIntegrity_CMWWaitBeforeGQ = C_CMW_WAIT_BEFORE_GQ;
//DebugN(dsConfigData, g_unSystemIntegrity_CMWCheckingDelay);
        g_iCMWNB_ITERATION_FOR_GQ=dsTemp[3];
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_DataCallback", dsConfigData, g_unSystemIntegrity_CMWCheckingDelay, g_unSystemIntegrity_CMWWaitBeforeGQ, g_iCMWNB_ITERATION_FOR_GQ);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_HandleCommand
/**
Purpose:
handle any systemIntegrity command.

@param sDpe1: string, input, dpe
@param command: int, input, the systemIntegrity command
@param sDpe2: string, input, dpe
@param parameters: dyn_string, input, the list of parameters of the systemIntegrity command

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters)
{
	dyn_string exceptionInfo, dsResult;
	int i, len =dynlen(parameters);
        string sResult;
  string sMessage;
	
	switch(command) {
		case UN_SYSTEMINTEGRITY_ADD: // add the _UnSystemAlarm DP
			for(i=1; i<=len; i++) {
				unSystemIntegrity_CMW_checking(parameters[i], true, true, exceptionInfo);
			}
			break;
		case UN_SYSTEMINTEGRITY_DELETE: // delete the _UnSystemAlarm DP
			for(i=1; i<=len; i++) {
				unSystemIntegrity_CMW_checking(parameters[i], false, false, exceptionInfo);
				// remove dp from the alert list of applicationDP if it is in
				_unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern+DS_pattern+parameters[i], false, exceptionInfo);
				// delete the _UnSystemAlarm dps
				if(dpExists(c_unSystemAlarm_dpPattern+DS_pattern+parameters[i]))
					dpDelete(c_unSystemAlarm_dpPattern+DS_pattern+parameters[i]);

				// remove dp from the alert list of applicationDP if it is in
				_unSystemIntegrity_modifyApplicationAlertList(c_unSystemAlarm_dpPattern+DS_Time_pattern+parameters[i], false, exceptionInfo);
				// delete the _UnSystemAlarm dps
				if(dpExists(c_unSystemAlarm_dpPattern+DS_Time_pattern+parameters[i]))
					dpDelete(c_unSystemAlarm_dpPattern+DS_Time_pattern+parameters[i]);
			}
			break;
		case UN_SYSTEMINTEGRITY_ENABLE: // enable the _UnSystemAlarm DP
			for(i=1; i<=len; i++) {
				if(dpExists(c_unSystemAlarm_dpPattern+DS_pattern+parameters[i]))
					unSystemIntegrity_CMW_checking(parameters[i], false, true, exceptionInfo);
			}
			break;
		case UN_SYSTEMINTEGRITY_DISABLE: // disable the _UnSystemAlarm DP
			for(i=1; i<=len; i++) {
				if(dpExists(c_unSystemAlarm_dpPattern+DS_pattern+parameters[i]))
					unSystemIntegrity_CMW_checking(parameters[i], false, false, exceptionInfo);
			}
			break;
		case UN_SYSTEMINTEGRITY_DIAGNOSTIC: // give the list of _UnSystemAlarm DP and the state
			dpSet(UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
					UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", g_unSystemIntegrity_CMW_DSList);
			break;
		case UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC:
		  dynAppend(dsResult, getCurrentTime());
		  
		  dynAppend(dsResult, "g_unSystemIntegrity_CMW_DSList");
		  len = dynlen(g_unSystemIntegrity_CMW_DSList);
		  for(i=1; i<=len; i++) {
			sResult = g_unSystemIntegrity_CMW_DSList[i];
			dynAppend(dsResult, sResult);
		  }
		  
		  dynAppend(dsResult, "g_unSystemIntegrity_CMW_DSThreadId");
		  len = dynlen(g_unSystemIntegrity_CMW_DSThreadId);
		  for(i=1; i<=len; i++) {
			sResult = g_unSystemIntegrity_CMW_DSThreadId[i];
			dynAppend(dsResult, sResult);
		  }
		  
		  dynAppend(dsResult, "g_unSystemIntegrity_CMWCheckingDelay");
		  dynAppend(dsResult, g_unSystemIntegrity_CMWCheckingDelay);
		  
		  len=mappinglen(g_munSystemIntegrity_TimeAlarm);
		  for(i=1;i<=len;i++)
		  {
			dynAppend(dsResult, "g_munSystemIntegrity_TimeAlarm["+mappingGetKey(g_munSystemIntegrity_TimeAlarm, i)+"]");
			dynAppend(dsResult, mappingGetValue(g_munSystemIntegrity_TimeAlarm, i));
		  }
		  
		  len=mappinglen(g_m_iCMWDriver_GQThreadId);
		  for(i=1;i<=len;i++)
		  {
			dynAppend(dsResult, "g_m_iCMWDriver_GQThreadId["+mappingGetKey(g_m_iCMWDriver_GQThreadId, i)+"] ");
			dynAppend(dsResult, mappingGetValue(g_m_iCMWDriver_GQThreadId, i));
		  }
		  
		  dynAppend(dsResult, "g_unSystemIntegrity_CMWWaitBeforeGQ");
		  dynAppend(dsResult, g_unSystemIntegrity_CMWWaitBeforeGQ);
		  
		  len=mappinglen(g_bunSystemIntegrity_driverStarted);
		  for(i=1;i<=len;i++)
		  {
			dynAppend(dsResult, "g_bunSystemIntegrity_driverStarted["+mappingGetKey(g_bunSystemIntegrity_driverStarted, i)+"]");
			dynAppend(dsResult, mappingGetValue(g_bunSystemIntegrity_driverStarted, i));
		  }
		  
		  len=mappinglen(g_m_iGQAllowed);
		  for(i=1;i<=len;i++)
		  {
			dynAppend(dsResult, "g_m_iGQAllowed["+mappingGetKey(g_m_iGQAllowed, i)+"]");
			dynAppend(dsResult, mappingGetValue(g_m_iGQAllowed, i));
		  }
		  
		  dynAppend(dsResult, "g_CMWSystemIntegrity_GeneralQueryThreadId");
		  dynAppend(dsResult, g_CMWSystemIntegrity_GeneralQueryThreadId);
		  
		  dynAppend(dsResult, "g_iCMWNB_ITERATION_FOR_GQ");
		  dynAppend(dsResult, g_iCMWNB_ITERATION_FOR_GQ);
		  
		  len = dynlen(g_CMW_diDriver);
		  for(i=1;i<=len;i++)
		  {
			dynAppend(dsResult, "g_CMW_diDriver["+i+"]");
			dynAppend(dsResult, g_CMW_diDriver[i]);
		  }

		  dpSet(UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
				UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", dsResult);
		  break;
		case UN_SYSTEMINTEGRITY_DEBUG:
		  if(dynlen(parameters) >= 2)
		  {
			g_b32DebugLevel = parameters[1];
			g_sDebugFilter = parameters[2];
		  }
		  else
		  {
			g_b32DebugLevel = 0;
			g_sDebugFilter = "";
		  }
		  dpSet(UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
				UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", makeDynString(g_b32DebugLevel, g_sDebugFilter));
		  break;
		case UN_SYSTEMINTEGRITY_VERSION:
		  sMessage = unGenericDpFunctions_getComponentsVersion(makeDynString("unCMWClient"));
		  dpSet(UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.commandResult", command, 
				UN_SYSTEM_INTEGRITY_CMW+UN_SYSTEMINTEGRITY_EXTENSION+".diagnostic.result", makeDynString(getCurrentTime(),sMessage));
		  break;
		default:
			break;
	}

	if(dynlen(exceptionInfo) > 0) {
		if(isFunctionDefined("unMessageText_sendException")) {
			unMessageText_sendException("*", "*", "unSystemIntegrity_CMW_HandleCommand", "user", "*", exceptionInfo);
		}
// handle any error uin case the send message failed
		if(dynlen(exceptionInfo) > 0) {
			DebugN(getCurrentTime(), exceptionInfo);
		}
	}

  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_HandleCommand", command, parameters);
// do here the dpConnect to the list of drivers
	if(g_bCMWNotStarted) {
		dpConnect("unSystemIntegrity_CMW_ConnectionDriverCB", CONNECTION_DRIVER);
		g_bCMWNotStarted = false;
	}
        if(g_CMWSystemIntegrity_GeneralQueryThreadId <= 0)
          g_CMWSystemIntegrity_GeneralQueryThreadId = startThread("unSystemIntegrity_CMWGeneralQueryInternalThread");

//DebugN(sDpe1, command, parameters, g_unSystemIntegrity_CMW_DSList);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_ConnectionDriverCB
/**
Purpose:
callback on the list of connected driver.

@param sDpe1: string, input, dpe
@param diDriverNum: dyn_int, input, the list of driver number started

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_ConnectionDriverCB(string sDp, dyn_int diDriverNum)
{
	dyn_string dsList = g_unSystemIntegrity_CMW_DSList;
	string sDpName;
	int i, len;

	len = dynlen(dsList);
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_ConnectionDriverCB GQ", diDriverNum, dsList);
	for(i=1;i<=len;i++) {
		sDpName = substr(dsList[i], strlen(c_unSystemAlarm_dpPattern+DS_pattern), strlen(dsList[i]));
//DebugTN("unSystemIntegrity_CMW_ConnectionDriverCB GQ", sDpName, drvNbr);
		unSystemIntegrity_CMW_ConnectionDriver(sDpName, diDriverNum);
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_ConnectionDriver
/**
Purpose:
Set the started state of the driver.

@param sDpName: string, input, device dp name
@param diDriverNum: dyn_int, input, the list of driver number started

Usage: Public

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_ConnectionDriver(string sDpName, dyn_int diDriverNum)
{
	int drvNbr;
	
	if(dpExists(sDpName)) {
		dpGet(sDpName+".communication.driver_num", drvNbr);
		if(dynContains(diDriverNum, drvNbr)>0) {
                  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_ConnectionDriver", sDpName, drvNbr);
// set the GQ bit so all data will bypass the smoothing
//DebugTN("drv", "_Driver" + drvNbr);
//				dpSet("_Driver" + drvNbr + ".SM:_original.._value", 1);
//				dpSet("_Driver" + drvNbr + ".GQ:_original.._value", 1);
			g_bunSystemIntegrity_driverStarted[drvNbr] = true;
		}
		else
			g_bunSystemIntegrity_driverStarted[drvNbr] = false;
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_checking
/**
Purpose:
This function register/de-register the callback funtion of CMW manager. This function can also create the _unSystemAlarm_CMW dp 
with the alarm config but it cannot delete it.

	@param sDp: string, input, data point name
	@param bCreate: bool, input, true create the dp and the alarm config if it does not exist, enable it
	@param bRegister: bool, input, true do a dpConnect, false do a dpDisconnect
	@param exceptionInfo: dyn_string, output, exception are returned here

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/

unSystemIntegrity_CMW_checking(string sDp, bool bCreate, bool bRegister, dyn_string &exceptionInfo)
{
	string sDpToChek, dp, dpTime, dpError, dpPlc, plcHostname, description, sDpCMWHW;
	int res, thId, pos, drvNum, len,i;
	bool bError = false;
	dyn_string plc_plc_list;

	string alertPanel,alertHelp;
	bool configExists, isActive;
	int alertConfigType;
	dyn_string alertTexts,alertClasses,summaryDpeList,alertPanelParameters;
	dyn_float alertLimits;
	
// remove the system name
	res = strpos(sDp, ":");
	if(res >= 0)
		dp = substr(sDp, res+1, strlen(sDp));
	else
		dp = sDp;
	sDpToChek = dp;
	dpTime = c_unSystemAlarm_dpPattern+DS_Time_pattern+dp;
	dp = c_unSystemAlarm_dpPattern+DS_pattern+dp;
//DebugN("unSystemIntegrity_CMW_checking", dp, bCreate, bRegister, sDpToChek, dpTime);

// in case the Dp is not of type _UnPlc do nothing. Should not happen, just for security.
	if(dpTypeName(sDpToChek) != UNCMW_DPTYPE)
		return;
	dpGet(sDpToChek+".communication.driver_num", drvNum);

	if(bCreate) {
// get the plc hostname and the driver number
		plcHostname = unGenericDpFunctions_getAlias(sDpToChek);
//		dpGet( sDpToChek+".communication.counter:_distrib.._driver", drvNum);
// create the sDp and its alarm config if it is not existing
		if(!dpExists(dp)) {
			description = getCatStr("unSystemIntegrity", "PLC_DS_DESCRIPTION")+plcHostname+" -> DS driver "+ drvNum;
			unSystemIntegrity_createSystemAlarm(sDpToChek, DS_pattern, description, exceptionInfo);
		}
// create the sDp time and its alarm config if it is not existing
		if(!dpExists(dpTime)) {
			description = getCatStr("unSystemIntegrity", "PLC_DS_TIME_DESCRIPTION")+plcHostname+" -> DS driver "+ drvNum;
			unSystemIntegrity_createSystemAlarm(sDpToChek, DS_Time_pattern, description, exceptionInfo);
		}
	}

	if(bRegister) {
		unSystemIntegrity_CMW_InitializeGQThread(sDpToChek);
// connect to the callback function if not already done
		pos = _unSystemIntegrity_isInList(g_unSystemIntegrity_CMW_DSList, dp);
		if(pos <= 0) {
                  unSystemIntegrity_addCMWGeneralQuery(drvNum);
	// add it in the list, because one of the callback function needs it.				
			pos = _unSystemIntegrity_setList(g_unSystemIntegrity_CMW_DSList, dp, true);

// start the thread for checking the counter
			thId = startThread(UN_SYSTEM_INTEGRITY_CMW_check, sDpToChek, drvNum);
			g_unSystemIntegrity_CMW_DSThreadId[pos] = thId;
			if(thId<0)
				bError = true;

			if(!bError) {
				g_munSystemIntegrity_TimeAlarm[dpSubStr(sDpToChek, DPSUB_SYS_DP)] = -1;
	// dpConnect to the stime of the counter
				res = dpConnect(c_unSystemIntegrity_CMW_Time_callback, 
																					sDpToChek+".communication.counter:_online.._stime_inv",
																					sDpToChek+".communication.counter:_online.._stime");
				if(res<0)
					bError = true;
				res = dpConnect("unSystemIntegrity_CMW_GQCB", sDpToChek+".configuration.GQAllowed", sDpToChek+".communication.driver_num");
				if(res<0)
					bError = true;
			}
			if(bError) {

				pos = _unSystemIntegrity_isInList(g_unSystemIntegrity_CMW_DSList, dp);
				if(pos > 0) {
		// kill the threads if they were started
					res = stopThread(g_unSystemIntegrity_CMW_DSThreadId[pos]);
// and then remove it from the list
					pos = _unSystemIntegrity_setList(g_unSystemIntegrity_CMW_DSList, dp, false);
//DebugN("rmv", res);
					dynRemove(g_unSystemIntegrity_CMW_DSThreadId, pos);
				}

				fwException_raise(exceptionInfo, "ERROR", 
							"unSystemIntegrity_CMW_checking():"+getCatStr("unSystemIntegrity", "CANNOT_DP_CONNECT") +dp,"");
			}
			else {
//DebugN("Enable", g_unSystemIntegrity_CMW_DSThreadId, "end");
			// set the enable to true and activate the alarm if it is not activated.
				dpSet(dp+".enabled", true);
				unAlarmConfig_mask(dp+".alarm", false, exceptionInfo);
				dpSet(dpTime+".enabled", true);
				unAlarmConfig_mask(dpTime+".alarm", false, exceptionInfo);
			// get all the PLC-PLC dp, and enable them
				plc_plc_list = dpNames(c_unSystemAlarm_dpPattern+PLC_PLC_pattern+sDpToChek+"*", c_unSystemAlarm_dpType);
				len = dynlen(plc_plc_list);
				for(i=1;i<=len;i++) {
					dpSet(plc_plc_list[i]+".enabled", true);
					unAlarmConfig_mask(plc_plc_list[i]+".alarm", false, exceptionInfo);
				}

			// get all the FESystemAlarm dp, and enable them, check the Ok state and set the correct value
				plc_plc_list = dpNames(c_unSystemAlarm_dpPattern+FE_pattern+sDpToChek+"*", c_unSystemAlarm_dpType);
				len = dynlen(plc_plc_list);
				for(i=1;i<=len;i++) {
					dpSet(plc_plc_list[i]+".enabled", true);
					unAlarmConfig_mask(plc_plc_list[i]+".alarm", false, exceptionInfo);
				}

// set the GQ bit so all data will bypass the smoothing
//DebugTN("add", "_Driver" + drvNum + ".SM:_original.._value");
//				dpSet("_Driver" + drvNum + ".SM:_original.._value", 1);
//				dpSet("_Driver" + drvNum + ".GQ:_original.._value", 1);
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_checking register", dp);
			}
		}
	}
	else {
// disconnect the callback function
//DebugN("Disable", g_unSystemIntegrity_CMW_DSThreadId, "end");
		pos = _unSystemIntegrity_isInList(g_unSystemIntegrity_CMW_DSList, dp);
		if(pos > 0) {
                  unSystemIntegrity_removeCMWGeneralQuery(drvNum);
// kill the thread
			res = stopThread(g_unSystemIntegrity_CMW_DSThreadId[pos]);
			if(res<0)
				bError = true;

			g_munSystemIntegrity_TimeAlarm[dpSubStr(sDpToChek, DPSUB_SYS_DP)] = -1;
// dpDisconnect stime of the counter
			res = dpDisconnect(c_unSystemIntegrity_CMW_Time_callback, 
																					sDpToChek+".communication.counter:_online.._stime_inv",
																					sDpToChek+".communication.counter:_online.._stime");
			if(res<0)
				bError = true;

			res = dpDisconnect("unSystemIntegrity_CMW_GQCB", sDpToChek+".configuration.GQAllowed", sDpToChek+".communication.driver_num");
			if(res<0)
				bError = true;

			if(bError) {
				fwException_raise(exceptionInfo, "ERROR", 
							"unSystemIntegrity_CMW_checking():"+getCatStr("unSystemIntegrity", "CANNOT_DP_DISCONNECT") +dp,"");
			}
			else {
			// remove from list
				pos = _unSystemIntegrity_setList(g_unSystemIntegrity_CMW_DSList, dp, false);
//DebugN("rmv", res);
				dynRemove(g_unSystemIntegrity_CMW_DSThreadId, pos);
			// set the enable to false and de-activate the alarm and acknowledge it if necessary
				dpSet(dp+".enabled", false, dp+".alarm", c_unSystemIntegrity_no_alarm_value);
				unAlarmConfig_mask(dp+".alarm", true, exceptionInfo);
				dpSet(dpTime+".enabled", false, dpTime+".alarm", c_unSystemIntegrity_no_alarm_value);
				unAlarmConfig_mask(dpTime+".alarm", true, exceptionInfo);
			// get all the PLC-PLC dp, and disable them
				plc_plc_list = dpNames(c_unSystemAlarm_dpPattern+PLC_PLC_pattern+sDpToChek+"*", c_unSystemAlarm_dpType);
				len = dynlen(plc_plc_list);
				for(i=1;i<=len;i++) {
// do not reset the alarm value because this DPE should be connected to an address config
					dpSet(plc_plc_list[i]+".enabled", false);
					unAlarmConfig_mask(plc_plc_list[i]+".alarm", true, exceptionInfo);
				}

			// get all the FESystemAlarm dp, and disable them, check the Ok state and set the correct value
				plc_plc_list = dpNames(c_unSystemAlarm_dpPattern+FE_pattern+sDpToChek+"*", c_unSystemAlarm_dpType);
				len = dynlen(plc_plc_list);
				for(i=1;i<=len;i++) {
// do not reset the alarm value because this DPE should be connected to an address config
					dpSet(plc_plc_list[i]+".enabled", false);
					unAlarmConfig_mask(plc_plc_list[i]+".alarm", true, exceptionInfo);
				}
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_checking de-register", dp);
			}
		}
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMWCheck
/**
Purpose:
This function checks if the counter of the CMW dp is modified periodically. If not it generates and alarm.

	@param sCMWDp: string, input, data point name

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMWCheck(string sCMWDp, int iDrvNum)
{
	int waitingTime, oldAlarmValue = -1;
	int oldValue = -1, newValue, alarmValue;
	int iTime, iOldTime=-1;
	bool bEnabled;
	string sModPlc;
	int res, iPlcNumber;
	
	while(true) {
	// if 0 set it to g_unSystemIntegrity_CMWCheckingDelay
		waitingTime = g_unSystemIntegrity_CMWCheckingDelay;
		if(waitingTime <= 0)
			waitingTime = c_unSystemIntegrity_defaultCMWCheckingDelay;
	// wait
		delay(waitingTime);
	
	// get the value of the counter and the timeout PLCChecking
		dpGet(sCMWDp+".communication.counter", newValue,
					sCMWDp+".communication.counter:_online.._stime", iTime);
//DebugN(newValue, oldValue, iTime, iOldTime);
		if((newValue == oldValue) && (iTime == iOldTime)) { // if smae value, check if same time
		// counter was not modified and same timestamp, set an alarm
			alarmValue = c_unSystemIntegrity_alarm_value_level1;
		}
		else	
		// counter was modified, reset the alarm
			alarmValue = c_unSystemIntegrity_no_alarm_value;

		if(oldAlarmValue != alarmValue) {
			dpSet(c_unSystemAlarm_dpPattern+DS_pattern+sCMWDp+".alarm", alarmValue);
			unSystemIntegrity_CMW_checkState(alarmValue, iDrvNum, g_unSystemIntegrity_CMWWaitBeforeGQ);
		}
		unSystemIntegrity_CMW_checkDriver(iDrvNum, g_unSystemIntegrity_CMWWaitBeforeGQ);

//DebugN("Check counter", sCMWDp, oldValue, newValue, alarmValue);	
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMWCheck", sCMWDp, oldValue, newValue, alarmValue);
		oldValue = newValue;
		oldAlarmValue = alarmValue;
		iOldTime = iTime;
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_checkState
/**
Purpose:
Initiate GQ if just started if no alarm.

	@param sDp1: string, input, data point name
	@param bStimeInvalid: bool, input, state of the stime bit of the counter

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_checkState(int alarmValue, int iDrvNum, int iDelay)
{
	if(alarmValue == c_unSystemIntegrity_no_alarm_value) {
// set the GQ bit so all data will bypass the smoothing
//DebugTN("check", "_Driver" + iDrvNum + ".SM:_original.._value");
//				dpSet("_Driver" + iDrvNum + ".SM:_original.._value", 1);
//				dpSet("_Driver" + iDrvNum + ".GQ:_original.._value", 1);
//DebugTN("Normal check", iDrvNum);
            unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_checkState", iDrvNum, iDelay);
		unSystemIntegrity_CMW_GQ(iDrvNum, iDelay);
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_checkDriver
/**
Purpose:
Initiate GQ if just started if the driver was just started.

	@param sDp1: string, input, data point name
	@param bStimeInvalid: bool, input, state of the stime bit of the counter

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_checkDriver(int iDrvNum, int iDelay)
{
	if(!mappingHasKey(g_bunSystemIntegrity_driverStarted,iDrvNum))
		g_bunSystemIntegrity_driverStarted[iDrvNum] = true;
	if(g_bunSystemIntegrity_driverStarted[iDrvNum]) {
//DebugTN("DRV startup", iDrvNum);
          unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_checkDriver", iDrvNum, iDelay);
		unSystemIntegrity_CMW_GQ(iDrvNum, iDelay);
		g_bunSystemIntegrity_driverStarted[iDrvNum] = false;
	}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_GQCB
/**
Purpose:
Callback function to if the GQ is allowed.

	@param sDp1: string, input, data point name
	@param bStimeInvalid: bool, input, state of the stime bit of the counter

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_GQCB(string sDp, bool bGQAllowed, string sDp2, int iDriver)
{
//DebugN(iDriver, bGQAllowed);
	g_m_iGQAllowed[iDriver] = bGQAllowed;
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMWTimeCallback
/**
Purpose:
Callback function to check if the time is correctly set in the PLC.

	@param sDp1: string, input, data point name
	@param bStimeInvalid: bool, input, state of the stime bit of the counter

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMWTimeCallback(string sDp1, bool bStimeInvalid, string sDp2, int iTime)
{
	string sDp, sKey = dpSubStr(sDp1, DPSUB_SYS_DP);
	int value, oldValue = -1;
	int iCurrentTime;
	bool bHasKey;
	
//DebugN("stime", sDp1, bStimeInvalid, g_munSystemIntegrity_TimeAlarm);

	if(mappingHasKey(g_munSystemIntegrity_TimeAlarm, sKey)) {
		oldValue = g_munSystemIntegrity_TimeAlarm[sKey];
		bHasKey = true;
	}
// get the dpname without the systemname
	sDp = dpSubStr(sDp1, DPSUB_DP);
// set the alarm 
	if(bStimeInvalid) { // bad time synchro
		value = c_unSystemIntegrity_alarm_value_level1;
	}
	else { // good time synchro
	// check if too late in the past
		iCurrentTime = getCurrentTime();
		if(iTime <= iCurrentTime) {
			if((iCurrentTime - iTime) > CONST_CHANGE_DIFF) { // too far in the past
//DebugN("alarm too late");
				value = 2*c_unSystemIntegrity_alarm_value_level1;
			}
			else
				value = c_unSystemIntegrity_no_alarm_value;
		}
		else
			value = c_unSystemIntegrity_no_alarm_value;
	}
	if(oldValue != value) {
		dpSet(c_unSystemAlarm_dpPattern+PLC_DS_Time_pattern+sDp+".alarm", value);
//DebugN("set", oldValue, value);
	}		
        unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMWTimeCallback", sKey, oldValue, value);
	oldValue = value;

	if(bHasKey)
		g_munSystemIntegrity_TimeAlarm[sKey] = value;
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMW_GeneralQuery
/**
Purpose:
Do a GW on the corresponding driver.

	@param sDp1: string, input, data point name
	@param bStimeInvalid: bool, input, state of the stime bit of the counter

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
	. PVSS version: 3.0 
	. operating system: W2000, NT and Linux, but tested only under W2000 and Linux.
	. distributed system: yes.
*/
unSystemIntegrity_CMW_GeneralQuery(string sSystemName, int iDriverNumber)
{
//DebugTN("_Driver"+iDriverNumber+".GQ");
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMW_GeneralQuery", sSystemName, iDriverNumber);
	dpSet(sSystemName+"_Driver"+iDriverNumber+".GQ", 0);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_CMWGeneralQueryInternalThread
/**
Purpose:
Do GQ on the list of configured drivers

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
  . PVSS version: 3.6SP2 
  . operating system: WXP and Linux
*/
unSystemIntegrity_CMWGeneralQueryInternalThread()
{
  int i, len;
  int iWaiting;
  dyn_int diDriver;
  bool bGQ;
  int iCMWNB_ITERATION_FOR_GQ=0;
  
  while(true)
  {
    iWaiting = g_unSystemIntegrity_CMWCheckingDelay;
    if(iWaiting <=0)
      iWaiting = c_unSystemIntegrity_defaultCMWCheckingDelay;
    delay(iWaiting);
    if((g_iCMWNB_ITERATION_FOR_GQ>0) && (iCMWNB_ITERATION_FOR_GQ >= g_iCMWNB_ITERATION_FOR_GQ))
    {
      iCMWNB_ITERATION_FOR_GQ = 0;
      synchronized(g_CMW_diDriver) {
        diDriver = g_CMW_diDriver;
      }
      dynUnique(diDriver);
      len = dynlen(diDriver);
      for(i=1;i<=len;i++)
      {
        if(mappingHasKey(g_m_iGQAllowed, diDriver[i])) 
          bGQ = g_m_iGQAllowed[diDriver[i]];
        else
          bGQ = false;
        if(bGQ)
          unSystemIntegrity_CMW_GeneralQuery("", diDriver[i]);
      }
    }
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_CMWGeneralQueryInternalThread", 
                           diDriver, bGQ, iWaiting, g_iCMWNB_ITERATION_FOR_GQ, iCMWNB_ITERATION_FOR_GQ);
    if(g_iCMWNB_ITERATION_FOR_GQ>0)
      iCMWNB_ITERATION_FOR_GQ++;
    else
      iCMWNB_ITERATION_FOR_GQ = 0;
  }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_addCMWGeneralQuery
/**
Purpose:
Add the driver to the list of driver to do GQ

  @param iDriverNum: int, input, the driver number

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
  . PVSS version: 3.6SP2 
  . operating system: WXP and Linux
*/
unSystemIntegrity_addCMWGeneralQuery(int iDriverNum) synchronized(g_CMW_diDriver)
{
  dynAppend(g_CMW_diDriver, iDriverNum);
  unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_addCMWGeneralQuery", 
                             iDriverNum);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_removeCMWGeneralQuery
/**
Purpose:
Remove the driver to the list of driver to do GQ

  @param iDriverNum: int, input, the driver number

Usage: Internal

PVSS manager usage: CTRL (WCCOACTRL)

Constraints:
  . PVSS version: 3.6SP2 
  . operating system: WXP and Linux
*/
unSystemIntegrity_removeCMWGeneralQuery(int iDriverNum) synchronized(g_CMW_diDriver)
{
  int iPos = dynContains(g_CMW_diDriver, iDriverNum);
  if(iPos > 0) {
    dynRemove(g_CMW_diDriver, iPos);
    unGenericDpFunctions_debugTN(DEBUG_ONLY_BIT0, "CMW systemIntegrity", "unSystemIntegrity_removeCMWGeneralQuery", 
                               iDriverNum);
  }
}

//------------------------------------------------------------------------------------------------------------------------

//@}

